
# coding: utf-8

# # Base Object

# In[6]:


import json
from collections.abc import Mapping


# In[7]:


class BaseObject( Mapping ):
    """
    Base object.
    """
    
    def __init__( self, properties, defaults ):
        """
        Creates a new BaseObject.
        
        :param properties: Initial property values. 
        :param defaults: Default property values.
        """
        self._properties = defaults.keys() # save white listed names
        
        properties = { **defaults, **properties } # set defaults
        for prop in self._properties:
            val = properties[ prop ]
            
            if ( prop is 'metadata' ) and isinstance( val, list ):
                md = {}
                
                # convert metadata list to dictionary
                for datum in val:
                    md_val = datum[ 'value' ]
                    
                    # cast value to correct type, comes as string
                    if datum[ 'type' ] is 'number':
                        try:
                            # check if int
                            md_val = int( md_val )

                        except ValueError as err:
                            # try as float
                            md_val = float( md_val )
                        
                    elif datum[ 'type' ] is 'json':
                        md_val = json.loads( md_val )
        
                    md[ datum[ 'name' ] ] = md_val
                
                val = md
                
            setattr( self, prop, val )
        
    
    @property
    def properties( self ):
        """
        :returns: List of valid properties.
        """
        return self._properties
    
    
    def __getitem__( self, item ):
        if item is '_properties':
            raise KeyError( item )
        
        return getattr( self, item )
        
    
    def __len__( self ):
        return len( self.__dict__ )
    

    def __iter__( self ):
        d = self.__dict__.copy()
        del d[ '_properties' ]
        
        yield from d
            
    
    def __json__( self ):
        """
        :returns: Dictionary to write to JSON.
        """
        return dict( self )


# In[8]:


class BaseObjectJSONEncoder( json.JSONEncoder ):
    """
    JSON encoder for Base Objects.
    """
    
    def default( self, obj ):
        if isinstance( obj, BaseObject ):
            return obj.__json__()
        
        else:
            # raise TypeError( 'Object {} is not a BaseObject.'.format( obj ) )
            super().default( obj )


# # Work




# coding: utf-8

# # Metadata

# In[ ]:


from .base_object import BaseObject


# In[ ]:


class Metadata( BaseObject ):
    """
    
    """
    
    def __init__( self ):
        """
        
        """
        pass





# coding: utf-8

# # Asset

# In[28]:


from .resource import Resource


# In[25]:


class Asset( Resource ):
    """
    An Asset.
    """
    
    def __init__( self, **kwargs ):
        """
        Creates a new Asset.
        
        :param **kwargs: Initial property values.
        """
        defaults = {
            'file':         None,
            'parent':       None,
            'creator':      None,
            'creator_type': None,
        }
        
        super().__init__( kwargs, defaults )





# coding: utf-8

# # Collection Interface

# In[1]:


from abc import ABC, abstractmethod


# In[ ]:


class CollectionInterface( ABC ):
    """
    Interface of required methods for a Collection.
    """
    
    
    @abstractmethod
    def find( self, search ):
        """
        Finds an object.
        
        :param search: Dictionary to filter objects.
        """
        pass
    
    
    @abstractmethod
    def find_one( self, search ):
        """
        Finds an object.
        
        :param search: Dictionary to filter objects.
        """
        pass
    
    @abstractmethod
    def insert_one( self, obj ):
        """
        Insert a single object.
        
        :param obj: Object to insert.
        """
        pass
    
    
    @abstractmethod
    def insert_many( self, objs ):
        """
        Insert a multiple objects.
        
        :param objs: List of object to insert.
        """
        pass
    
    
    @abstractmethod
    def replace_one( self, search, obj ):
        """
        Replace a single object.
        
        :param search: Dictionary to filter objects.
        :param obj: Object to insert.
        """
        pass
    
    
    @abstractmethod
    def update_one( self, search, update ):
        """
        Update a single object.
        
        :param search: Dictionary to filter objects.
        :param update: Dictionary to update.
        """
        pass
    
    
    @abstractmethod
    def update_many( self, search, update ):
        """
        Update multiple objects.
        
        :param search: Dictionary to filter objects.
        :param update: Dictionary to update.
        """
        pass
    
    
    @abstractmethod
    def delete_one( self, search, update ):
        """
        Delete a single object.
        
        :param search: Dictionary to filter objects.
        """
        pass
    
    
    @abstractmethod
    def delete_many( self, search, update ):
        """
        Delete multiple objects.
        
        :param search: Dictionary to filter objects.
        """
        pass





# coding: utf-8

# # Driver interface

# In[3]:


from abc import ABC, abstractmethod


# In[4]:


class DriverInterface( ABC ):
    """
    Interface for drivers.
    """
    
    @abstractmethod
    def output_asset( self, asset ):
        """
        Modifies asset coming from the database, going to the user.
        
        :param asset: Assets to modify.
        :returns: Modified asset.
        """
        pass
    
    
    @abstractmethod
    def output_assets( self, assets ):
        """
        Modifies Assets coming from the database, going to the user.
        
        :param assets: List of Assets to modify.
        :returns: List of modified Assets.
        """
        pass
    
    
    @abstractmethod
    def input_asset( self, asset ):
        """
        Modifies an asset coming from the user, going to the database.
        
        :param asset: Asset to be modified.
        :returns: The modified asset.
        """
        pass
    
    
    @abstractmethod
    def get_script( self, script ):
        """
        Gets the specified script if needed, returning the local path to it.
        
        :param script: Path to the script.
        :returns: Local path.
        """
        pass





# coding: utf-8

# # Script

# In[ ]:


from .base_object import BaseObject
from .resource import Resource


# In[ ]:


class Script( Resource ):
    """
    A Script.
    """
    
    def __init__( self, **kwargs ):
        """
        Creates a new Asset.
        
        :param **kwargs: Initial property values.
        """
        defaults = {
            'file':         None,
            'language':     None,
            'description':  None,
            'version':      0
        }
        
        super().__init__( kwargs, defaults )


# In[ ]:


class ScriptAssociation( BaseObject ):
    """
    A ScriptAssociation
    """
    
    def __init__( self, **kwargs ):
        """
        :param **kwargs: Initial property values.
        """
        defaults = {
            'script': None,
            'priority': 0,
            'autorun':  True
        }
        
        super().__init__( kwargs, defaults )
        
        
    def __lt__( self, other ):
        """
        Defines comparison based on priority.
        
        :param other: ScriptAssociation for comparison.
        :returns: True if priority is less than other, False otherwise.
        """
        return ( self.priority < other.priority )
        





# coding: utf-8

# # Role

# In[1]:


from .base_object import BaseObject


# In[ ]:


class Role( BaseObject ):
    """
    Represents a role.
    """
    
    def __init__( self, **kwargs ):
        """
        :params **kwargs: Initial parameters
        """
        defaults = {
            '_id':    None, # remove _id field
            'user': None,
            'role': None
        }
        
        super().__init__( kwargs, defaults )





# coding: utf-8

# # Container

# In[ ]:


from .resource import Resource
from .script import ScriptAssociation
from .role import Role


# In[ ]:


class Container( Resource ):
    """
    A Container.
    """
    
    def __init__( self, **kwargs ):
        """
        Creates a new Container.
        
        :param **kwargs: Initial property values.
        """
        defaults = {
            'parent':       None,
            'children':     [],
            'assets':       [],
            'scripts':      []
        }
        
        super().__init__( kwargs, defaults )
        
        # roles
        for index, role in enumerate( self.roles ):
            self.roles[ index ] = Role( **role )
    
        # scripts
        for index, script in enumerate( self.scripts ):
            self.scripts[ index ] = ScriptAssociation( **script )
    
    
    
#     def find_descendants( self, search = {}, level = None ):
#         """
#         Gets descendents.
        
#         :param match: Dictionary filter or None for no filter. [Default: None]
#         :param level:
#         :returns: List of Containers.
#         """
#         pass
    
    
#     def find_assets( self, search = {} ):
#         """
        
#         """
#         pass
    
    
#     def find_scritps( self, search = {} ):
#         """
        
#         """
#         pass





# coding: utf-8

# # Runner Core

# In[ ]:


import os
import sys
import json
import subprocess
from concurrent.futures import ThreadPoolExecutor

from .classes.container import Container


# In[ ]:


class Runner:
    """
    For performing analysis on a Thot project.
    
    Hooks are:
        + get_container: Method to retrieve the Container being run. 
            Provided the root id, should return the Container.
            [Siganture: ( root ) => ( Container )]
            [Required]
            
        + get_script_info: Method to retrieve require information about the Script being run.
            [Signature: ( ScriptAssociation.script ) => ( script id, script path ) ]
            [Required]
        
        + script_error: Runs when a Script creates an error.
            [Signature: ( error, ignore_errors ) => ()]
            
        + assets_added: Run after a Script analysis is complete, passed ids of the added Assets.
            [Signature: ( added_assets ) => ()]
            
        + complete: Run after a subtree completes.
            [Signature: () => ()]
    """
    
    def __init__( self ):
        """
        Initializes a new Runner.
        """
        self.hooks = {
            'get_container':   None,
            'get_script_info': None,
            'script_error':    self._default_script_error_handler,
            'assets_added':    None,
            'complete':        None
        }
    
    
    def register( self, hook, method ):
        """
        Registers a hook method.
        
        :param hook: Name of the hook to register.
        :param method: Method to run.
        """
        if hook not in self.hooks:
            # hook is invalid,
            # all hook names are defined in constructor
            raise ValueError( 'Invalid hook name.' )
        
        self.hooks[ hook ] = method
    
    
    
    def run_script( self, script_id, script_path, container_id ):
        """
        Runs the given program form the given Container.

        :param script_id: Id of the script.
        :param script_path: Path to the script.
        :param container: Id of the container to run from.
        :returns: Script output. Used for collecting added assets.
        """
        # setup environment
        env = os.environ.copy()
        env[ 'THOT_CONTAINER_ID' ] = container_id # set root container to be used by thot library
        env[ 'THOT_SCRIPT_ID' ]    = script_id    # used in project for adding Assets

        # TODO [0]: Ensure safely run
        # run program
        try:
            return subprocess.check_output(
                'python {}'.format( script_path ),
                shell = True,
                env = env
            )
        
        except subprocess.CalledProcessError as err:
            err.cmd = '[{}] '.format( container_id ) + err.cmd
            raise err
    
    
    # TODO [2]: Allow running between certain depths.
    def eval_tree( 
        self,
        root, 
        scripts = None,
        ignore_errors = False, 
        multithread = False, 
        verbose = False,
    ):
        """
        Runs scripts on the Container tree.
        Uses DFS, running from bottom up.

        :param root: Container.
        :param scripts: List of scripts to run, or None for all. [Default: None]
        :param ignore_errors: Continue running if an error is encountered. [Default: False]
        :param multithread: Evaluate tree using multiple threads. 
            True to use default number of threads, or an integer 
            to specify how many threads to use. 
            If interpreted as boolean is False, will use a single thread.
            CAUTION: May decrease runtime, but also locks system and can not kill.
            [Default: False] [Default Threads: 5]
        :param verbose: Print evaluation information. [Default: False]
        """
        self._check_hooks()
        
        root = self.hooks[ 'get_container' ]( root )
        if not isinstance( root, Container ):
            root = Container( **root )

        # eval children
        kwargs = {
            'scripts': scripts,
            'ignore_errors': ignore_errors,
            'multithread': multithread,
            'verbose': verbose
        }
        
        if multithread:
            max_workers = (
                5 # default
                if multithread is True else
                multithread
            )
            
            with ThreadPoolExecutor( max_workers = max_workers ) as executer:
                # only enable threading at top level
                kwargs[ 'multithread' ] = False
                
                executer.map( 
                    lambda child: self.eval_tree( child, **kwargs ), 
                    root.children 
                )

        else:
            for child in root.children:
                # recurse
                self.eval_tree( child, **kwargs )

        # TODO [1]: Check filtering works for local projects.
        # filter scripts to run
        root.scripts.sort()
        run_scripts = (
            root.scripts
            if scripts is None else
            filter( lambda assoc: assoc.script in scripts, root.scripts ) # filter scripts
        )

        # eval self
        added_assets = []
        for association in run_scripts:
            if not association.autorun:
                continue

            ( script_id, script_path ) = self.hooks[ 'get_script_info' ]( association.script )

            if verbose:
                print( 'Running script {} on container {}'.format( script_id, root._id )  )

            try:
                script_assets = self.run_script( 
                    str( script_id ), # convert ids if necessary
                    script_path, 
                    str( root._id ) 
                ) 

            except Exception as err:
                self.hooks[ 'script_error' ]( err, ignore_errors )

            if self.hooks[ 'assets_added' ]:
                script_assets = [ 
                    json.loads( asset ) for asset
                    in script_assets.decode().split( '\n' )
                    if asset
                ]

                self.hooks[ 'assets_added' ]( script_assets )


        if self.hooks[ 'complete' ]:
            self.hooks[ 'complete' ]()
            
            
    def _check_hooks( self ):
        """
        Verifies registered hooks.
        
        :throws: Error if registered hook is invalid.
        """
        if not self.hooks[ 'get_container' ]:
            raise Error( 'Required hook get_container is not set.' )
        
        if not self.hooks[ 'get_script_info' ]:
            raise Error( 'Required hook get_script_info is not set.' )
    
    
    @staticmethod
    def _default_script_error_handler( err, ignore_errors = False ):
        """
        
        """
        if ignore_errors:
            # TODO [2]: Only return errors after final exit.
            # collect errors for output at end
            print( '[{}] {}'.format( root._id, err ) )

        else:
            raise err


# # Work




# coding: utf-8

# In[ ]:


import random
from datetime import datetime
from json import JSONEncoder

from .base_object import BaseObject
from .role import Role
from .note import Note
from .metadata import Metadata


# In[ ]:


class Resource( BaseObject ):
    
    def __init__( self, properties, defaults ):
        """
        Creates a new Resource.
        
        :param properties: Initial property values. 
        :param defaults: Default property values.
        """       
        resource_defaults = {
            '_id':          None,
            'created':      datetime.now(),
            'type':         None,
            'name':         None,
            'description':  None,
            'roles':        [],
            'tags':         [],
            'notes':        [],
            'metadata':     {}
        }
        
        # merge defaults
        defaults = { **resource_defaults, **defaults }
        
        super().__init__( properties, defaults )
        
        
    def __json__( self ):
        """
        Returns a dictionary to write to JSON.
        """        
        obj = dict( self )
        
        if isinstance( obj[ 'created' ], datetime ):
            # convert datetime to string
            time_format = '%Y-%m-%dT%H:%M:%S%Z' # yyyy-mm-ddThh:mm:ss+/-<UTC offset>
            obj[ 'created' ] = obj[ 'created' ].strftime( time_format )
        
        return obj


# In[ ]:


class ResourceJSONEncoder( JSONEncoder ):
    """
    JSON encoder for Base Objects.
    """
    
    def default( self, obj ):
        if isinstance( obj, Resource ):
            return obj.__json__()
        
        else:
            super().default()





# coding: utf-8

# # Thot Interface

# In[3]:


import os
from abc import ABC, abstractmethod


# In[4]:


class ThotInterface( ABC ):
    
    def __init__( self, root = None ):
        """
        :param root: Root Container. 
            If None. attempts to get root from
            environment variable THOT_CONTAINER_ID.
            [Default: None]
        :raise ValueError: If neither root nor THOT_CONTAINER_ID are not set.
            
        """
        if root is None:
            try:
                root = os.environ[ 'THOT_CONTAINER_ID' ]
                
            except KeyError as err:
                raise ValueError( 'Root Container not passed and THOT_CONTAINER_ID not set.' )
        
        self._root = root
        self._db = None
    
    
    @property
    def root( self ):
        return self._root
    
    
    @abstractmethod
    def find_container( self, search = {} ):
        """
        Gets a single Container matching search criteria.
        
        :param search: Dictionary of search criteria.
            [Default: {}]
        :returns: Container matching search criteria or None.
        """
        return self._db.containers.find_one( search )
    
    
    @abstractmethod
    def find_containers( self, search = {} ):
        """
        Gets Containers matching search criteria.
        
        :param search: Dictionary of search criteria.
            [Default: {}]
        :returns: List of Containers matching search.
        """
        return self._db.containers.find( search )
    
    
    @abstractmethod
    def find_asset( self, search = {} ):
        """
        Gets a single Asset matching search criteria.
        
        :param search: Dictionary of search criteria.
            [Default: {}]
        :returns: Asset matching search criteria or None.
        """
        return self._db.assets.find_one( search )
    
    
    @abstractmethod
    def find_assets( self, search = {} ):
        """
        Gets Assets matching search criteria.
        
        :param search: Dictionary of search criteria.
            [Default: {}]
        :returns: List of Assets matching search.
        """
        return self._db.assets.find( search )
    
    
    @abstractmethod
    def add_asset( self, asset ):
        """
        Adds a new Asset.
        
        :param asset: Asset to create.
        """
        pass
    
    
    @staticmethod
    def dev_mode():
        """
        Whether the script is being run in development mode or not.
        
        :returns: False if being run from the runner, True otherwise.
        """
        return False




import setuptools

with open( 'README.rst', 'r' ) as f:
	long_desc = f.read()

classifiers = [
	"Programming Language :: Python :: 3",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: OS Independent",
]

project_urls = {
	'Documentation': 	'https://thot-data-docs.readthedocs.io/',
	'Source Code': 		'https://github.com/bicarlsen/thot-core',
	'Bug Tracker':		'https://github.com/bicarlsen/thot-core/issues'
}


setuptools.setup(
	name='thot-core',
	version = '0.4.0',
	author='Brian Carlsen',
	author_email = 'carlsen.bri@gmail.com',
	description = 'Thot data analysis and management.',
	long_description = long_desc,
	long_description_content_type = 'text/x-rst',
	url = 'http://www.thot-data.com',
	packages = setuptools.find_packages(),
	project_urls = project_urls,
	classifiers = classifiers,
	
	install_requires = [],

	package_data = {},

	entry_points = {
		'console_scripts': []
	}
)